import numpy as np
import pytest
import scanpy as sc
import scipy.sparse as spr

import triku as tk


@pytest.fixture
def getpbmc3k():
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_genes(adata, min_cells=10)
    sc.pp.filter_cells(adata, min_genes=10)
    sc.pp.neighbors(adata)

    # tk.tl.triku(adata)

    return adata


@pytest.mark.exception_check
def test_triku_check_count_mat_20000_vars():
    adata = sc.datasets.blobs(
        n_variables=22000, n_centers=3, cluster_std=1, n_observations=500
    )
    adata.X = np.abs(adata.X).astype(int)
    print(adata.X)

    sc.pp.filter_cells(adata, min_genes=10)
    sc.pp.filter_genes(adata, min_cells=1)

    sc.pp.pca(adata)
    sc.pp.neighbors(adata)

    assert adata.X.shape[1] > 20000

    tk.tl.triku(adata)


@pytest.mark.exception_check
def test_triku_check_nonaccepted_type():
    adata = sc.datasets.blobs(
        n_variables=2000, n_centers=3, cluster_std=1, n_observations=500
    )
    adata.X = np.abs(adata.X).astype(int)
    adata.var_names = [0] + list(np.arange(len(adata.var_names) - 1))

    try:
        tk.tl.triku(adata.X)
    except BaseException:
        pass
    else:
        raise BaseException


@pytest.mark.exception_check
def test_triku_dense_sparse_matrices(getpbmc3k):
    adata_dense = getpbmc3k
    adata_sparse = getpbmc3k.copy()

    adata_dense.X = adata_dense.X.toarray()
    adata_sparse.X = spr.csr.csr_matrix(adata_sparse.X)

    tk.tl.triku(adata_sparse)
    tk.tl.triku(adata_dense)

    for key in ["knn", "n_features", "s", "n_windows", "min_knn", "dist_conn"]:
        assert (
            adata_sparse.uns["triku_params"][None][key]
            == adata_dense.uns["triku_params"][None][key]
        )

    assert (
        adata_sparse.uns["triku_params"][None]["knn_array"]
        != adata_dense.uns["triku_params"][None]["knn_array"]
    ).nnz == 0

    assert np.all(
        adata_dense.var["triku_distance"] == adata_sparse.var["triku_distance"]
    )


@pytest.mark.exception_check
def test_triku_dense_sparse_matrices_raw(getpbmc3k):
    adata_dense = getpbmc3k
    adata_sparse = getpbmc3k.copy()

    adata_dense.X = adata_dense.X.toarray()
    adata_sparse.X = spr.csr.csr_matrix(adata_sparse.X)

    adata_dense.raw = adata_dense
    adata_sparse.raw = adata_sparse

    tk.tl.triku(adata_sparse, use_raw=True)
    tk.tl.triku(adata_dense, use_raw=True)

    for key in ["knn", "n_features", "s", "n_windows", "min_knn", "dist_conn"]:
        assert (
            adata_sparse.uns["triku_params"][None][key]
            == adata_dense.uns["triku_params"][None][key]
        )

    assert (
        adata_sparse.uns["triku_params"][None]["knn_array"]
        != adata_dense.uns["triku_params"][None]["knn_array"]
    ).nnz == 0

    assert np.all(
        adata_dense.var["triku_distance"] == adata_sparse.var["triku_distance"]
    )


@pytest.mark.exception_check
def test_check_count_mat_negative(getpbmc3k):
    adata = getpbmc3k

    adata.X[0, 0] = -1

    try:
        tk.tl.triku(adata)
        error = 1
    except BaseException:
        error = 0

    if error == 1:
        raise


@pytest.mark.exception_check
def test_check_zero_counts():
    adata = sc.datasets.pbmc3k()
    sc.pp.neighbors(adata)

    try:
        tk.tl.triku(
            adata
        )  # There should be zeros and raise an error -> therefore it should yield error = 0
        error = 1
    except BaseException:
        error = 0

    if error == 1:
        raise

    adata_nonzero = sc.datasets.pbmc3k()
    sc.pp.filter_genes(adata_nonzero, min_counts=1)
    sc.pp.neighbors(adata_nonzero)

    tk.tl.triku(adata_nonzero)



import io
import logging


class TqdmToLogger(io.StringIO):
    # Adapted from https://gitlhub.com/tqdm/tqdm/issues/313
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    def __init__(self, logger, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO
        self.buf = ""

    def write(self, buf):
        self.buf = buf.strip("\r\n\t ")

    def flush(self):
        self.logger.log(self.level, self.buf)

    # to create the bar:
    # tqdm_out = TqdmToLogger(logger, level=logging.INFO)
    # for x in tqdm(range, file=tqdm_out)



import logging


"""
We will add a new deggubing level, TRIKU, which is halfway between DEBUG and INFO.
Ray uses INFO to produce its output, which is considerable, and we trust it. To avoid it, we will create an intermediate
level, TRIKU. The default level will be INFO, but for debugging purposses we will use TRIKU.
"""
TRIKU_LEVEL = (logging.INFO + logging.DEBUG) // 2
logging.addLevelName(TRIKU_LEVEL, "TRIKU")

triku_logger = logging.getLogger(__name__)
logger_handler = logging.StreamHandler()  # Handler for the logger
logger_handler.setFormatter(
    logging.Formatter("%(levelname)s | %(asctime)s - %(name)s - %(message)s\n")
)
triku_logger.addHandler(logger_handler)
triku_logger.setLevel(logging.INFO)



import numpy as np
import scipy.sparse as spr

from triku.logg import TRIKU_LEVEL
from triku.logg import triku_logger


def check_count_mat(mat: spr.csr.csr_matrix):
    """
    This function outputs a warning if we suspect the matrix is in logarithm value
    """
    triku_logger.info("Checking integrity of matrix.")

    if np.min(mat) < 0:
        error_msg = "The count matrix contains negative values. Triku is supposed to run with positive matrices."

        triku_logger.error(error_msg)
        raise BaseException(error_msg)

    if mat.shape[1] > 20000:
        triku_logger.warning(
            "The count matrix contains more than 25000 genes. We recommend filtering some genes, up to "
            "15000 - 18000 genes. You can do that in scanpy with the function 'sc.pp.filter_genes()'."
        )


def check_null_genes(arr_counts: np.ndarray):
    """
    Removes columns (genes) that have zero sum. These genes interfere in the analysis and are not useful at all.
    """
    triku_logger.info("Checking zero-count genes.")

    if np.any(np.array(arr_counts.sum(0))[0] == 0):
        error_msg = (
            "There are genes with no counts. Remove those genes first."
            "You can use sc.pp.filter_genes(adata, min_cells=5)."
        )

        triku_logger.error(error_msg)
        raise BaseException(error_msg)


def return_arr_counts(object_triku, get_from_raw=None):
    triku_logger.log(TRIKU_LEVEL, "Obtaining count matrix and gene list.")
    # Check type of object and return the matrix as corresponded

    if get_from_raw:
        triku_logger.info(
            "Using raw matrix. If you want to use the current matrix, set use_raw=False (although "
            "we discourage it)."
        )
        arr_counts = object_triku.raw.X
    else:
        arr_counts = object_triku.X

    if not spr.isspmatrix(arr_counts):
        triku_logger.warning(
            "X is dense. We will set the matrix to sparse format (csr_matrix)."
        )
        arr_counts = spr.csr_matrix(
            arr_counts
        )  # We will require the array of counts to be sparse

        if get_from_raw:
            triku_logger.warning(
                "Can't set adata.raw.X to sparse. You can do it before setting the raw as adata.X = scipy.sparse.csr.csr_matrix(adata.X)"
            )
        else:
            object_triku.X = arr_counts

    triku_logger.log(
        TRIKU_LEVEL, "Array of counts\n{arr_counts}, shape:{arr_counts.shape}",
    )
    triku_logger.log(TRIKU_LEVEL, "Array of genes\n{arr_genes}")
    return arr_counts


def get_arr_counts(object_triku, use_raw):
    # Process the use_raw argument, and return the matrix of cells and genes accordingly.
    if use_raw:
        if object_triku.raw is not None:
            arr_counts = return_arr_counts(object_triku, get_from_raw=True)
        else:
            arr_counts = return_arr_counts(object_triku, get_from_raw=False)

    else:
        arr_counts = return_arr_counts(object_triku, get_from_raw=False)

    check_count_mat(arr_counts)
    check_null_genes(arr_counts)

    return arr_counts



# import gc
from typing import Tuple

import bottleneck
import numpy as np
import scipy.sparse as spr
import scipy.stats as sts
from scipy.signal import fftconvolve

from triku.logg import TRIKU_LEVEL
from triku.logg import triku_logger

# import logging
# from tqdm import tqdm
# from triku.genutils import TqdmToLogger


def return_knn_array(object_triku, dist_conn, knn):
    # Distances array contains a pairwise relationship between cells, based on distance.
    # We will binarize that array to select equally all components with non-zero distance.
    # We finally add the identity matrix to select as neighbour the own cell.
    try:
        # Dist array shows indices of neighbors, but not the own cell.
        # To add the own cell we use spr.identity.

        if dist_conn == "dist":
            knn_array = (
                object_triku.obsp[  # type:ignore
                    "distances"
                ]
                > 0
            ) + spr.identity(len(object_triku)).astype(
                bool
            )  # Saves memory

        elif dist_conn == "conn":
            # Connectivities adds an extra layer of neighbours that I don't know how it happens. https://github.com/theislab/scanpy/issues/1984
            # The problem with connectivities it that nonzero elements are not always knn, but it changes.
            # In the dense version of triku we selected the knn indexes with highest connectivities. To do that in sparse, we will do it row by row.
            # To make the process faster we construct the array from zero by setting indptr, indexes and data.
            # 1) Indptr are 0, knn, 2*knn, 3*knn, 4*knn, ...
            # 2) Data are 1 throughout the array.
            # 3) The indices of the best connectivities are selected via argsort of the best values (selected in object_triku.obsp["connectivities"].data)

            knn_array_conn_indices = np.ones(
                len(object_triku) * knn, dtype=int
            )
            knn_array_conn_data = np.ones(len(object_triku) * knn, dtype=bool)
            knn_array_conn_indptr = np.arange(
                0, (len(object_triku) + 1) * knn, knn
            )

            for row in range(len(object_triku)):
                start_counts, end_counts = (
                    object_triku.obsp["connectivities"].indptr[row],
                    object_triku.obsp["connectivities"].indptr[row + 1],
                )
                data_row = object_triku.obsp["connectivities"].data[
                    start_counts:end_counts
                ]
                idx_row = object_triku.obsp["connectivities"].indices[
                    start_counts:end_counts
                ]

                if (
                    len(data_row) >= knn
                ):  # bottleneck works much faster, but it requires that the number of connected features be >= knn.
                    # If not (because there are some), it fails. Then, it goes to the classical argsort, which is slower but effective.
                    argidx = bottleneck.argpartition(-data_row, knn - 1)
                    idx_pos = idx_row[
                        argidx[: knn - 1]
                    ]  # knn - 1 because the own index is not within the best
                else:  # Sometimes the number of neighbors, based on connectivities, is smaller, and it fails. We fallback to distances in that case. The number of
                    # cases is very small in that case.
                    idx_pos = object_triku.obsp["distances"].indices[
                        object_triku.obsp["distances"]
                        .indptr[row] : object_triku.obsp["distances"]
                        .indptr[row + 1]
                    ]

                knn_array_conn_indices[row * knn] = row
                knn_array_conn_indices[
                    row * knn + 1 : (row + 1) * knn
                ] = idx_pos

            knn_array = spr.csr.csr_matrix(
                (
                    knn_array_conn_data,
                    knn_array_conn_indices,
                    knn_array_conn_indptr,
                ),
                shape=(len(object_triku), len(object_triku)),
            )
        else:
            raise SyntaxError(
                "dist_conn can only take 'dist' and 'conn' values."
            )

    except KeyError:
        triku_logger.warning(
            """Deprecation issue. Recent versions of scanpy save knn matrices in .obsp.
        In future versions of triku we will select these matrices from .obsp exclusively."""
        )

        knn_array = (
            object_triku.uns[  # type:ignore
                "neighbors"
            ][
                "connectivities"
            ]
            > 0
        ) + spr.identity(len(object_triku)).astype(
            bool
        )  # Saves memory

    return knn_array


def get_n_divisions(arr_counts: spr.csr.csr_matrix) -> int:
    diff = np.abs(
        (arr_counts - arr_counts.floor()).sum()
    )  # Faster .floor() than X.astype(int) (up to x2 for large arrays)
    triku_logger.log(
        TRIKU_LEVEL, f"Difference between int and float array is {diff}"
    )

    if diff < 1:
        n_divisions = 1
    else:
        n_divisions = 15  # Arbitrarily chosen. Prime/large numbers are better because 2, 3, 4 yield strange results during binning.

    triku_logger.log(TRIKU_LEVEL, f"Number of divisions set to {n_divisions}")
    return n_divisions


def return_knn_expression(
    arr_expression: spr.csr.csr_matrix, knn_indices: spr.csr.csr_matrix
) -> spr.csr.csr_matrix:
    """
    This function returns an array with the knn expression per gene and cell. To calculate the expression per gene
    we are going to apply the dot product of the neighbor indices and the expression.

    That is, if we have n_g as number of genes, and n_c as number of cells, the matrix product would be:

        Mask          Expr         Result
    (n_c x n_c) · (n_c x n_g) = (n_c x n_g)

    Then, the Result matrix would have in each cell, the summed expression of that gene in the knn (and also the
    own cell).

    In this step we do not mask the array. Previously, after the calculation of the expression, we masked the knn expression of the
    cells that were originally expressing that gene. That is, for any gene, the knn expression of the cells are were originally
    not expressing that gene was set to zero. We do that because we saw that not doing that produced "dirtier" EMD calculations.
    The thing is that since the matrices are csr, constructing a masked array requires a new matrix and selecting the elements from
    the knn matrix, or deleting the existing ones based on the count array, in both cases time consuming.

    To save that time, we will simply in the convolution step select the expression values with the mask for each gene, because that
    selection has to be done anyways.
    """

    knn_expression = knn_indices.dot(arr_expression)

    triku_logger.log(
        TRIKU_LEVEL,
        "knn_expression: {knn_expression} | {knn_expression.shape}",
    )

    return knn_expression


def compute_conv_idx(
    counts_gene: np.ndarray, knn: int, p_zeros: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Given a GENE x CELL matrix, and an index to select from, calculates the convolution of reads for that gene index.
    The function returns the
    """
    y_probs = np.bincount(counts_gene) / len(
        counts_gene
    )  # Important to transform count to probabilities
    # to keep the convolution constant.
    # THIS y_probs MUST have the probs of P(X=0) because the
    # random neighbors might have 0 counts!!!  [TODO: check this!!!]

    # We are calculating the convolution of cells with positive expression. Thus, in the first distribution
    # we have to remove the cells with 0 reads, and rescale the probabilities.
    y_probs_0 = np.bincount(counts_gene) / len(counts_gene)
    y_probs = y_probs_0.copy() * (1 - p_zeros)
    y_probs[0] = p_zeros

    """
    Reynolds
    len(y_probs)        scp.fftconvolve     scp.convolve     np.convolve
    6783                52                  52.5             2000
    5129                41                  43               933
    3072                23.5                23.6             334
    1783                14.5                15.3             129
    1202                10.3                13.4             59.4
    669                 6.6                 22.2             20.8
    298                 4.19                5.32             5.18
    237                 3.74                3.48             3.23
    188                 3.54                2.98             2.78
    159                 3.11                1.95             1.63
    117                 2.93                1.41             1.12
    52                  2.33                0.8              0.63
    34                  2.77                0.75             0.41
    19                  1.98                0.61             0.23
    13                  1.94                0.45             0.27
    8                   1.83                0.55             0.19
    PBMC 10k
    len(y_probs)        scp.fftconvolve     scp.convolve     np.convolve
    4123                31.4                32.1             604
    1324                11.1                12.1             45.5
    885                 7.98                15.1             33.8
    725                 6.98                13.1             14.9
    551                 5.8                 15               14.4
    345                 4.3                 7.1              6.7
    255                 3.6                 3.9              3.7
    190                 3.3                 2.7              2.3
    119                 2.7                 1.7              1.5
    70                  2.7                 0.9              0.84
    36                  2.1                 0.6              0.3
    20                  2.1                 0.5              0.2
    """

    if (
        len(y_probs) > 250
    ):  # This is important. For some genes, if the counts are too big, np.convolve crashes!!
        func = fftconvolve
    else:
        func = np.convolve

    arr_convolve = func(
        y_probs_0, y_probs
    )  # First iteration always with itself

    for _ in range(knn - 1):
        arr_convolve = func(arr_convolve, y_probs,)

    # Important because some convolutions yield negative close-to-zero values that break emd
    arr_convolve[np.isclose(arr_convolve, 0)] = 0

    arr_prob = (
        arr_convolve / arr_convolve.sum()
    )  # This is just in case the sum is bigger than 1

    x_conv, y_conv = np.arange(len(arr_prob)), arr_prob

    return x_conv, y_conv


def calculate_emd(
    knn_counts: np.ndarray,
    x_conv: np.ndarray,
    y_conv: np.ndarray,
    n_divisions: int,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns "normalized" earth movers distance (EMD). The function calculates the x positions and probabilities
    of the "real" dataset using the knn_counts, and the x positions and probabilities of the convolution as attributes.

    To normalize the distance, it is divided by the standard deviation of the convolution. Since the convolution
    is already given as a distribution, mean and variance have to be calculated "by hand".
    """
    dist_range = np.arange(np.max(knn_counts) + 1)
    # np.bincount transforms [3, 3, 4, 1, 2, 9] into [0, 1, 1, 2, 1, 0, 0, 0, 0, 1]
    real_vals = np.bincount(knn_counts) / len(knn_counts)

    # IMPORTANT: either for std or emd calculation, all x variables must be scaled back!
    dist_range = dist_range / n_divisions
    x_conv = x_conv / n_divisions

    emd = sts.wasserstein_distance(dist_range, x_conv, real_vals, y_conv)

    mean = (x_conv * y_conv).sum()
    std = np.sqrt((y_conv * (x_conv - mean) ** 2).sum())

    return x_conv, emd / std


def compute_convolution_and_emd(
    array_counts_csc: spr.csc.csc_matrix,
    array_knn_counts_csc: spr.csc.csc_matrix,
    idx: int,
    knn: int,
    min_knn: int,
    n_divisions: int,
) -> np.ndarray:
    """Calculate the convolution and emd given the array with counts and with knn counts.
    To do the convolution we will select the gene column from each array.

    From the array of counts we will simply select the values, and from the array of knn counts we will select the values
    of the indices from the array of counts (arr_counts[:, idx].indices).

    Then, we are going to make the array integer. To do that, we recall the n_divisions argument, that applies
    binning to the unit. For instance, if the expression of a gene is 5.23 and 5 bins are set, the new expression is
    int(5.23 * 5) = int(26.15) = 26 -> 26 / 5 = 5.2 (so we lose 0.03 of expression).
    This is a scaling step

    To do sparse array accession faster we will play with csr_matrix.data, csr_matrix.indptr and csr_matrix.indices attributes.
    This makes the code a bit obscure, but makes the selection faster (or at least guarantees it is not slower).
    """

    # If you don't see this part well, trust it, or see https://stackoverflow.com/questions/52299420/scipy-csr-matrix-understand-indptr
    # array_counts_csc.indptr[beggining:end] yields the location of array_counts_csc.indices of the column of interest (beginning).
    start_counts, end_counts = (
        array_counts_csc.indptr[idx],
        array_counts_csc.indptr[idx + 1],
    )
    start_knn, end_knn = (
        array_knn_counts_csc.indptr[idx],
        array_knn_counts_csc.indptr[idx + 1],
    )

    # array_counts_csc.indices[beggining:end] yields the location of array_counts_csc.data of the column of interest (beginning).
    indices_counts = array_counts_csc.indices[start_counts:end_counts]
    indices_counts_knn = array_knn_counts_csc.indices[start_knn:end_knn]
    # because indices_counts_knn array has the indices from indices_counts + extra, this line extracts the overlapping indices.
    bool_mask = np.in1d(indices_counts_knn, indices_counts, assume_unique=True)

    counts_gene = array_counts_csc.data[start_counts:end_counts]
    knn_counts = array_knn_counts_csc.data[start_knn:end_knn][bool_mask]

    """
    The next step is to multiply the counts and knn_counts by n_divisions, and set it to int because otherwise the convolution fails.
    """

    counts_gene = (counts_gene * n_divisions).astype(int)
    knn_counts = (knn_counts * n_divisions).astype(int)

    # This part is necessary for compute_conv_idx, because we don't have zeros in counts_gene, and we need them to create the probability
    # distribution including zeros.
    p_zeros = 1 + (start_counts - end_counts) / array_counts_csc.shape[0]

    if len(counts_gene) < min_knn:
        emd = 0
    else:
        x_conv, y_conv = compute_conv_idx(counts_gene, knn, p_zeros)
        x_conv, emd = calculate_emd(knn_counts, x_conv, y_conv, n_divisions)

    return emd


def emd_calculation(
    array_counts_csc: spr.csr.csr_matrix,
    array_knn_counts_csc: spr.csr.csr_matrix,
    knn: int,
    min_knn: int,
    n_divisions: int,
) -> Tuple[list, list, np.ndarray]:
    """
    Calculation of convolution for each gene, and its emd. To do that we call compute_convolution_and_emd which,
    in turn, calls compute_conv_idx to calculate the convolution of the reads; and calculate_emd, to calculate the
    emd between the convolution and the knn_counts.

    Since we are working with counts of each gene, instead of each cell, we will get the csc forms of array_knn_counts and array_counts.
    This conversion takes some time and memory, but it does save a lot of time afterwards, when doing the column indexing.
    e.g. with a 50000 x 10000 matrix, doing csr -> csc and csc indexing takes 8s, whereas doing csr indexing takes 30 mins!!

    To make things faster we use ray parallelization. Ray selects the counts and knn counts on each gene, and computes
    the convolution and distance. The output result is, for each gene, the convolution distribution
    (x, and probabilities), and the distances.
    """
    n_genes = array_counts_csc.shape[1]

    triku_logger.log(TRIKU_LEVEL, "Running EMD calulation.")

    list_emd = [
        compute_convolution_and_emd(
            array_counts_csc,
            array_knn_counts_csc,
            idx_gene,
            knn,
            min_knn,
            n_divisions,
        )
        for idx_gene in range(n_genes)
    ]

    return np.array(list_emd)


def subtract_median(x, y, n_windows):
    """When working with EMD, we want to find genes with more deviation on emd compared with other genes with similar
    mean expression. With higher expressions EMD tends to increase. To reduce that basal level we will subtract the
    median EMD to the genes using a number of windows. The approach is quite reliable between 15 and 80 windows.

    Too many windows can over-normalize, and lose genes that have high emd but are alone in that window."""

    # We have to take the distance in logarithm to account for the wide expression ranges
    linspace = 10 ** np.linspace(
        np.min(np.log10(x)), np.max(np.log10(x)), n_windows + 1
    )
    y_adjust = y.copy()

    y_median_array = np.zeros(len(y))
    for i in range(n_windows):
        mask = (x >= linspace[i]) & (x <= linspace[i + 1])
        y_median_array[mask] = np.median(y[mask])

    y_adjust -= y_median_array

    return y_adjust


def get_cutoff_curve(y, s) -> float:
    """
    Plots a curve, and finds the best point by joining the extremes of the curve with a line, and selecting
    the point from the curve with the greatest distance.
    The distance between a point in a curve, and the straight line is set by the following equation
    if u,v is the point in the curve, and y = mx + b is the line, then
    x_opt = (u - mb + mv) / (1 + m^2)

    Here y attribute refers to the emd distances (after median subtraction preferably). Those distances are sorted,
    and ordered, and the curve is extracted from there.
    """

    min_y, max_y = np.min(y), np.max(y)
    m, b = (max_y - min_y) / len(y), min_y

    list_d = []

    for u, v in enumerate(np.sort(y)):
        x_opt = (u - m * b + m * v) / (1 + m ** 2)
        y_opt = x_opt * m + b
        d = (x_opt - u) ** 2 + (y_opt - v) ** 2

        list_d.append(d)

    # S is a corrector factor. It leverages the best value in the curve, and selects a more or less stringent
    # value in the curve. the maximum distance is multiplied by (1 - S), and the leftmost or rightmost index
    # is selected

    dist_s = (1 - np.abs(s)) * np.max(list_d)
    s_idx = np.argwhere(list_d >= dist_s)

    if s >= 0:
        max_d_idx = np.max(s_idx)
    else:
        max_d_idx = np.min(s_idx)

    return np.sort(y)[max_d_idx]



import time

import numpy as np
import pytest
import scanpy as sc
from scipy.sparse import csr_matrix

import triku as tk


selected_markers = [
    "CD79A",
    "CD14",
    "C3",
    "GZMB",
    "HMOX1",
    "ICAM4",
    "ITGA2B",
    "CLU",
]


@pytest.mark.general
def test_run_defaults():
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_cells(adata, min_genes=10)
    sc.pp.filter_genes(adata, min_cells=10)
    sc.pp.pca(adata)
    sc.pp.neighbors(adata)
    t = time.time()
    tk.tl.triku(adata)
    print("TT", time.time() - t)
    print(adata.var.loc[selected_markers])
    for pos_gene in selected_markers:
        print(pos_gene)
        assert adata.var["highly_variable"].loc[pos_gene]


@pytest.mark.general
def test_run_defaults_random():
    # Generate random neigbors to set Wasserstein distance to 0
    adata = sc.datasets.blobs(
        n_variables=5000, n_centers=5, cluster_std=2.0, n_observations=1000
    )
    adata.X = np.abs(adata.X)
    sc.pp.filter_cells(adata, min_genes=30)
    sc.pp.filter_genes(adata, min_cells=30)
    sc.pp.pca(adata)
    sc.pp.neighbors(adata, n_neighbors=15)

    adata_random = adata.copy()

    knn_random_mat = np.zeros((len(adata), len(adata)))
    for row in range(len(adata)):
        mat_options = list(np.arange(row)) + list(
            np.arange(row + 1, (len(adata)))
        )
        mat_idx = np.random.choice(mat_options, 15)
        knn_random_mat[row, mat_idx] = 1
    knn_random_mat = csr_matrix(knn_random_mat)

    adata_random.obsp["distances"], adata_random.obsp["connectivities"] = (
        knn_random_mat,
        knn_random_mat,
    )
    tk.tl.triku(adata)
    tk.tl.triku(adata_random)

    assert adata_random.var["triku_distance_uncorrected"].mean() < 0.1
    assert adata.var["triku_distance_uncorrected"].mean() > 1.5


@pytest.mark.end
def test_end():
    pass



import logging

from ..logg import TRIKU_LEVEL
from ..logg import triku_logger


def set_level_logger(level):
    dict_levels = {
        "debug": logging.DEBUG,
        "triku": TRIKU_LEVEL,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }

    triku_logger.setLevel(dict_levels[level])



import numpy as np
import pytest
import scanpy as sc

import triku as tk


@pytest.fixture
def getpbmc3k():
    adata = sc.datasets.pbmc3k()
    sc.pp.filter_genes(adata, min_cells=10)
    sc.pp.filter_cells(adata, min_genes=10)
    sc.pp.neighbors(adata)

    # tk.tl.triku(adata)

    return adata


@pytest.mark.calc_check
def test_n_features(getpbmc3k):
    adata = getpbmc3k
    for n_feats in [1, 50, 100, 500, 1000]:
        tk.tl.triku(adata, n_features=n_feats)
        assert np.sum(adata.var["highly_variable"].values) == n_feats


@pytest.mark.output_check
def test_use_raw(getpbmc3k):
    adata = getpbmc3k
    tk.tl.triku(adata, use_raw=False)
    emd_not_raw = adata.var["triku_distance"]

    adata.raw = adata
    sc.pp.log1p(adata)
    tk.tl.triku(adata, use_raw=True)
    emd_raw = adata.var["triku_distance"]

    assert np.all(emd_raw.values == emd_not_raw.values)


@pytest.mark.output_check
def test_s(getpbmc3k):
    n_feats = []

    for s in [
        0.1,
        0.05,
        0,
        -0.05,
        -0.1,
    ]:  # This s order guarantees that the number of selected feats will be increasing
        adata = getpbmc3k
        tk.tl.triku(adata, s=s)
        n_feats.append(adata.var["highly_variable"].values.sum())

    assert sorted(n_feats) == n_feats


@pytest.mark.var_check
def test_n_divisions_check(getpbmc3k):
    adata = getpbmc3k
    assert np.sum(adata.X - adata.X.astype(int)) == 0

    adata.X = adata.X.astype(int)
    tk.tl.triku(adata)
    assert adata.uns["triku_params"][None]["n_divisions"] == 1

    adata.X = adata.X.astype(float)
    tk.tl.triku(adata)
    assert adata.uns["triku_params"][None]["n_divisions"] == 1

    sc.pp.log1p(adata)
    tk.tl.triku(adata)
    assert adata.uns["triku_params"][None]["n_divisions"] > 1


@pytest.mark.var_check
def test_names(getpbmc3k):
    adata = getpbmc3k

    tk.tl.triku(adata)
    tk.tl.triku(adata, name="sample")

    for var in [
        "triku_distance",
        "triku_distance_sample",
        "triku_distance_uncorrected",
        "triku_distance_uncorrected_sample",
    ]:
        assert var in adata.var

    assert None in adata.uns["triku_params"]
    assert "sample" in adata.uns["triku_params"]

    for key in ["knn", "n_features", "s", "n_windows", "min_knn", "dist_conn"]:
        assert (
            adata.uns["triku_params"][None][key]
            == adata.uns["triku_params"]["sample"][key]
        )

    assert (
        adata.uns["triku_params"][None]["knn_array"]
        != adata.uns["triku_params"]["sample"]["knn_array"]
    ).nnz == 0

    assert np.all(
        adata.var["triku_distance"] == adata.var["triku_distance_sample"]
    )
    assert np.all(
        adata.var["triku_distance_uncorrected"]
        == adata.var["triku_distance_uncorrected_sample"]
    )


@pytest.mark.output_check
def test_dist_conn(getpbmc3k):
    adata_dist = getpbmc3k.copy()
    tk.tl.triku(adata_dist, use_raw=False, dist_conn="dist")

    adata_conn = getpbmc3k.copy()
    tk.tl.triku(adata_conn, use_raw=False, dist_conn="conn")

    try:
        tk.tl.triku(adata_conn, use_raw=False, dist_conn="other")
        error = 0
    except SyntaxError:
        error = 1

    assert error == 1



import sys

if sys.version_info < (3,):
    sys.exit("triku requires Python >= 3.7")
from pathlib import Path

from setuptools import setup, find_packages


__author__ = ", ".join(["Alex M. Ascensión"])
__email__ = ", ".join(
    [
        "alexmascension@gmail.com",
        # We don’t need all, the main authors are sufficient.
    ]
)
__version__ = "2.1.1"


setup(
    name="triku",
    version=__version__,
    description="Feature selection method for Single Cell data.",
    long_description=Path("README.md").read_text("utf-8"),
    url="https://gitlab.com/alexmascension/triku",
    author=__author__,
    author_email=__email__,
    license="BSD",
    python_requires=">=3.7",
    install_requires=[
        module.strip()
        for module in Path("requirements.txt").read_text("utf-8").splitlines()
    ],
    packages=find_packages(),
    # `package_data` does NOT work for source distributions!!!
    # you also need MANIFTEST.in
    # https://stackoverflow.com/questions/7522250/how-to-include-package-data-with-setuptools-distribute
    # package_data={'': '*.txt'},
    # include_package_data=True,
    entry_points=dict(),
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



# import logging
import warnings
from typing import Union

import numpy as np
import scanpy as sc

from ..logg import TRIKU_LEVEL
from ..logg import triku_logger
from ..utils._general_utils import set_level_logger
from ..utils._triku_tl_utils import get_arr_counts
from ._triku_functions import emd_calculation
from ._triku_functions import get_cutoff_curve
from ._triku_functions import get_n_divisions
from ._triku_functions import return_knn_array
from ._triku_functions import return_knn_expression
from ._triku_functions import subtract_median

warnings.filterwarnings("ignore")  # To ignore Numba warnings


def triku(
    object_triku: Union[sc.AnnData, str],
    n_features: Union[None, int] = None,
    use_raw: bool = True,
    n_divisions: Union[None, int] = None,
    s: Union[int, float] = -0.01,
    n_windows: int = 75,
    min_knn: int = 6,
    name: Union[str, None] = None,
    verbose: Union[None, str] = "warning",
    dist_conn: Union[str] = "dist",
) -> dict:  # type:ignore
    """
    This function calls the triku method using python directly. This function expects an
    annData object or a csv / txt matrix of n_cells x n_genes. The function then returns an array / list
    of the selected genes.

    Parameters
    ----------
    object_triku : scanpy.AnnData or pandas.DataFrame
        Object with count matrix. If `pandas.DataFrame`, rows are cells and columns are genes.
    n_features : int
        Number of features to select. If None, the number is chosen automatically.
    use_raw : bool
        If True, selects the adata.raw, if it exists.
        To set the .raw propety, set as: adata.raw = adata.
        This matrix is adjusted to select the genes and cells that
        appear in the current adata. E.g. if we are running triku with a subpopulation, triku will select the cells
        from adata.raw of that subpopulation. If certain genes have been removed after saving the raw, triku will not
        consider the removed genes.
    n_divisions : int, None
        If the array of counts is not integer, number of bins in which each unit will be divided to account for
        that effect. For example, if n_divisions is 10, then 0.12 and 0.13 would be in the same bin, and 0.12 and 0.34
        in two different bins. If n_divisions is 2, the two cases would be in the same bin.
        The higher the number of divisions the more precise the calculation of distances will be. It will be more
        time consuming, though. If n_divisions is None, we will adjust it automatically.
    s : float
        Correction factor for automatic feature selection. Negative values imply a selction of more genes, and
        positive values imply a selection of fewer genes. We recommend values between -0.1 and 0.1.
    n_windows : int
        Number of windows used for median subtraction of Wasserstein distance.
    min_knn : int
        minimum number of expressed cells based on the knn to apply the convolution. If a gene has less than min_knn
        expressing cells, Wasserstein distance is set to 0, and the convolution is set as the knn expression.
    name: str
        Name of the run. If None, stores results in "triku_X". Else, stores it in "triku_X_{name}".
    dist_conn: str
        Uses adata.obsp["distances"] or adata.obsp["connectivities"] for knn array construction. From empirical analysis,
        "conn" shows slightly better results, but is slower.
    verbose : str ['debug', 'triku', 'info', 'warning', 'error', 'critical']
        Logger verbosity output.
    Returns
    -------
    list_features : list
        list of selected features
    """

    # Basic checks of variables and assertions!!!
    set_level_logger(verbose)

    for var in [
        n_features,
        n_windows,
        n_divisions,
    ]:
        assert (var is None) | (
            isinstance(var, int)
        ), f"The variable value {var} must be an integer!"

    if name is None:
        name_str = ""
    else:
        name_str = f"_{name}"

    # Check that neighbors are calculated. Else make the user calculate them!!!
    error_ms = "Neighbors not found in adata. Run sc.pp.neighbors() first."
    try:
        if "neighbors" not in object_triku.uns:  # type:ignore
            raise IndexError(error_ms)
    except AttributeError:
        raise IndexError(error_ms)

    # Assert that adata.X is sparse (warning to transform) and assert that gene names are unique.
    arr_counts = get_arr_counts(object_triku, use_raw=use_raw)

    # Get n_divisions if None:
    if n_divisions is None:
        n_divisions = get_n_divisions(arr_counts)

    """
    First step is to get the kNN for the expression matrix from the annData.
    The expected matrix from this step is a matrix of cells x kNN, where each column includes the neighbor index
    of the cell number. The matrix is sparse, so we will work with methods that adapt work with sparse matrices.
    """

    knn = object_triku.uns["neighbors"]["params"]["n_neighbors"]  # type:ignore

    # Boolean array showing the neighbors (including its own)
    knn_array = return_knn_array(object_triku, dist_conn, knn)

    # Calculate the expression in the kNN (+ own cell) for all genes [CAUTION! This array is unmasked!!!! (more explained inside the funcion)]
    triku_logger.info("Calculating knn expression")
    arr_knn_expression = return_knn_expression(arr_counts, knn_array)

    # Set to csc matrix. This allows a faster selection by column, where the genes are located. Accession times improve up to 1000x!!!
    array_counts_csc = arr_counts.tocsc()
    array_knn_counts_csc = arr_knn_expression.tocsc()

    # Apply the convolution, and calculate the EMD. The convolution is quite fast, but we will still paralellize it.
    triku_logger.info("EMD calculation")
    triku_logger.log(TRIKU_LEVEL, "min_knn set to {}".format(min_knn))
    array_emd = emd_calculation(
        array_counts_csc=array_counts_csc,
        array_knn_counts_csc=array_knn_counts_csc,
        knn=knn,
        min_knn=min_knn,
        n_divisions=n_divisions,
    )

    triku_logger.info("Subtracting median")
    mean_counts = arr_counts.mean(0).A[0]

    array_emd_subt_median = subtract_median(
        x=mean_counts, y=array_emd, n_windows=n_windows
    )

    # Selection of best genes, either by the curve method or as the N highest ones.
    if n_features is None:
        triku_logger.info("Selecting cutoff point")
        dist_cutoff = get_cutoff_curve(y=array_emd_subt_median, s=s)
    else:
        dist_cutoff = np.sort(array_emd_subt_median)[-(n_features + 1)]
    triku_logger.info("Cutoff point set to {dist_cutoff}")

    # Returns phase. Return if object is not an adata or if return is set to true.
    # We set it like that so that scanpy knows which are the hvg.
    hvg = array_emd_subt_median > dist_cutoff
    object_triku.var[  # type:ignore
        "highly_variable"
    ] = hvg

    object_triku.var[  # type:ignore
        f"triku_distance{name_str}"
    ] = array_emd_subt_median

    object_triku.var[  # type:ignore
        f"triku_distance_uncorrected{name_str}"
    ] = array_emd

    object_triku.var[  # type:ignore
        f"triku_highly_variable{name_str}"
    ] = hvg

    if "triku_params" not in object_triku.uns:  # type:ignore
        object_triku.uns["triku_params"] = {}  # type:ignore

    object_triku.uns["triku_params"][name] = {  # type:ignore
        "knn": knn,
        "n_features": n_features,
        "s": s,
        "n_windows": n_windows,
        "min_knn": min_knn,
        "n_divisions": n_divisions,
        "dist_conn": dist_conn,
        "knn_array": knn_array,
    }
